from .basic_utils import *
import re
import time
import shutil
import string
import random
Import("tqdm",globals())
Import("requests",globals())

def RandString(length:int, charset:str=string.ascii_uppercase + string.digits):
    """Return a random string.

    Args:
        length (int): Length of the random string.
        charset (str): The alphabet. Notice that if a word appears multiple times, the probability that it is chosen also multiples.
    Returns:
        str: A random string of length `length`, with all characters chosen from `charset`.
    """
    return ''.join(random.choice(charset) for _ in range(length))

def LineToFloats(line:str):
    """Read all float numbers from a line of text.

    Args:
        line (str): A line of text.
    Returns:
        List: All float numbers find in the line.
    """
    return [float(s) for s in re.findall(r"(?<!\w)[-+]?\d*\.?\d+(?!\d)",line)]

def FlattenList(L:List):
    """Flatten an embedded list to a 1D list.

    Args:
        L: The embedded list to be flattened.
    Returns:
        List: The flattened list.
    """
    F = lambda x:[e for i in x for e in F(i)] if isinstance(x,list) else [x]; return F(L)

def IP():
    """Return IP address string of the current device.

    Args:
        None
    Returns:
        str: The IP address string of the current device.
    """
    return requests.get('https://api.ipify.org').text

def FORMATTED_TIME(format="[%Y-%m-%d_%H.%M.%S]"):
    """Return the formatted current time.

    Args:
        format (str): The time format.
    Returns:
        str: The formatted current time.
    """
    return time.strftime(format,time.localtime(time.time()))

def Clipped(value, l, u):
    """Return a clipped value between lower bound `l` and upper bound `u` (closed).

    Args:
        value: A value.
        l: The lower bound.
        u: The upper bound.
    Returns:
        Any: The clipped value.
    """
    return min(max(value,l),u)

def Unique(ls:List, **sort_args):
    """Return a sorted list of unique elements.

    Args:
        ls (List): The input list (or other iterable).
        sort_args: Args for calling `sorted` on the subpaths in format of (path, full_path), only works if `ordered` is True. Notice that the sort function will only be applied at the base level instead of applied recursively. PPlease refer to function `BUILTIN_LISTPATHS_SORT_CRITERIA()` for built-in criteria.

    Returns:
        List: The sorted list of unique elements.
    """
    return sorted(list(set(ls)), **sort_args)

def TQDM(obj, **args):
    """Return a tqdm pbar, both iterable and int are supported for obj.

    Args:
        ls (List): The object to be tracked using tqdm.
        args: args for calling `tqdm.tqdm` on the object (or range).

    Returns:
        tqdm.tqdm: The tqdm pbar.
    """
    if 'dynamic_ncols' not in args:
        args['dynamic_ncols'] = True
    if isinstance(obj, int):
        return tqdm.tqdm(range(obj), **args)
    else:
        return tqdm.tqdm(obj, **args)

def ZIP(src, dst):
    """Zip a directory `src` to a `.zip` file.

    Args:
        src: The object to be tracked using tqdm.
        args: args for calling `tqdm.tqdm` on the object (or range).

    Returns:
        tqdm.tqdm: The tqdm pbar.
    """
    shutil.make_archive(Prefix(p2s(dst)), 'zip', p2s(src))


def PrintFile(file_handle, *args, **kwargs):
    """Print to a file_handle with auto-flush.

    Args:
        *args, **kwargs: arguments used in `print`.
    Returns:
        None
    """
    if 'flush' not in kwargs:
        kwargs['flush'] = True
    print(args, file=file_handle, **kwargs)


def PrintConsole(*args, **kwargs):
    """Print to stdout with auto-flush.

    Args:
        *args, **kwargs: arguments used in `print`.
    Returns:
        None
    """
    if 'flush' not in kwargs:
        kwargs['flush'] = True
    print(args, file=sys.stdout, **kwargs)


def PrintError(*args, **kwargs):
    """Print to stderr with auto-flush.

    Args:
        *args, **kwargs: arguments used in `print`.
    Returns:
        None
    """
    if 'flush' not in kwargs:
        kwargs['flush'] = True
    print(args, file=sys.stderr, **kwargs)

def NORMAL(obj):
    """Format an object into string of normal color in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return str(obj)


def ERROR(obj):
    """Format an object into string of error color (red) in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;31m' + str(obj) + '\x1b[0m'


def SUCCESS(obj):
    """Format an object into string of success color (green) in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;32m' + str(obj) + '\x1b[0m'


def WARNING(obj):
    """Format an object into string of warning color (yellow) in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;33m' + str(obj) + '\x1b[0m'


def COLOR1(obj):
    """Format an object into string of highlight color 1 (blue) in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;34m' + str(obj) + '\x1b[0m'


def COLOR2(obj):
    """Format an object into string of highlight color 2 (magenta) in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;35m' + str(obj) + '\x1b[0m'


def COLOR3(obj):
    """Format an object into string of highlight color 3 (cyan) in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;36m' + str(obj) + '\x1b[0m'


def BLACK(obj):
    """Format an object into string of black color in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;30m' + str(obj) + '\x1b[0m'


def RED(obj):
    """Format an object into string of red color in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;31m' + str(obj) + '\x1b[0m'


def GREEN(obj):
    """Format an object into string of green color in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;32m' + str(obj) + '\x1b[0m'


def YELLOW(obj):
    """Format an object into string of yellow color in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;33m' + str(obj) + '\x1b[0m'


def BLUE(obj):
    """Format an object into string of blue color in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;34m' + str(obj) + '\x1b[0m'


def MAGENTA(obj):
    """Format an object into string of magenta color in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;35m' + str(obj) + '\x1b[0m'


def CYAN(obj):
    """Format an object into string of cyan color in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;36m' + str(obj) + '\x1b[0m'


def WHITE(obj):
    """Format an object into string of white color in console.

    Args:
        obj: the object to be formatted.

    Returns:
        None
    """
    return '\x1b[1;37m' + str(obj) + '\x1b[0m'


from .file_utils import *
import io
import torch
import numpy as np
import torch.nn as nn
import torch.nn.init as nninit
from torch.utils.data import Dataset, Subset, ConcatDataset, DataLoader

def SaveTorch(obj, path, pickle_protocol:Optional[int]=None, _use_new_zipfile_serialization:bool=True):
    """Save an object as torch file.

    Args:
        obj: The object to be saved.
        path: The save path.
        pickle_protocol (int/None): The `pickle_protocol` argument for `torch.save`.
        _use_new_zipfile_serialization (int/None): The `_use_new_zipfile_serialization` argument for `torch.save`.
    Returns:
        None
    """
    CreateFile(path); path = p2s(path)
    if pickle_protocol is not None:
        torch.save(obj, path, pickle_protocol=pickle_protocol, _use_new_zipfile_serialization=_use_new_zipfile_serialization)
    else:
        torch.save(obj, path, _use_new_zipfile_serialization=_use_new_zipfile_serialization)

def LoadTorch(path, map_location=None):
    """Load an object from existing torch file.

    Args:
        path: The load path.
        map_location: The `map_location` argument for `torch.load`.
    Returns:
        Any: The loaded object.
    """
    assert (ExistFile(path)), (f"Path '{path}' does not exist!"); path = p2s(path)
    return torch.load(path, map_location=map_location)

def DumpsTorch(obj, pickle_protocol:Optional[int]=None, _use_new_zipfile_serialization:bool=True):
    """Save an object as torch bytearray.

    Args:
        obj: The object to be saved.
        pickle_protocol (int/None): The `pickle_protocol` argument for `torch.save`.
        _use_new_zipfile_serialization (int/None): The `_use_new_zipfile_serialization` argument for `torch.save`.
    Returns:
        bytearray: The torch bytearray.
    """
    buff = io.BytesIO()
    if pickle_protocol is not None:
        torch.save(obj, buff, pickle_protocol=pickle_protocol, _use_new_zipfile_serialization=_use_new_zipfile_serialization)
    else:
        torch.save(obj, buff, _use_new_zipfile_serialization=_use_new_zipfile_serialization)
    buff.seek(0); return buff.read()

class HeavenDataset(Dataset):
    def __init__(self, dataset, name="HeavenDataset"):
        self.dataset = dataset; self.name=name; self.__index__ = 0
    
    def __getitem__(self, i):
        if isinstance(i, int):
            return self.dataset[i]
        else:
            indices = np.array([j for j in range(len(self.dataset))])[i]
            return HeavenDataset(Subset(self, indices.tolist()),name=self.name)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]
    
    def __next__(self):
        if self.__index__>=len(self):
            raise StopIteration
        else:
            data = self[self.__index__]; self.__index__ += 1; return data

    def rename(self, name="HeavenDataset"):
        self.name = name
    
    def __len__(self):
        return len(self.dataset)
    
    def __or__(self, b):
        return HeavenDataset(ConcatDataset([self,b]),name=self.name+(","+b.name if hasattr(b,"name") else ""))

    def __str__(self):
        return ("{"+self.name+"}" if ',' in self.name else self.name)+f"(len={len(self)})"

class HeavenDataLoader:
    def __init__(self, dataloader, dataset=None, eternal=False):
        self.dataset = dataset
        self.dataloader = dataloader
        self.iterator = iter(dataloader)
        self.eternal = eternal

    def __iter__(self):
        for data in self.dataloader:
            yield data
        while self.eternal:
            for data in self.dataloader:
                yield data

    def __next__(self):
        try:
            data = next(self.iterator)
        except StopIteration:
            if self.eternal:
                self.iterator = iter(self.dataloader)
                data = next(self.iterator)
            else:
                raise StopIteration
        return data

    def __len__(self):
        return len(self.dataloader)

Import("timm.create_model@create_model",globals())
Import("torchvision.models._utils@TV_U",globals())

class TimmBackbone(nn.Module):
    def __init__(self,
        model_type="",
        input_layer_getter=lambda n:n.conv1,
        modify_input_channels=None,
        output_layer_name="fc",
        embedding_dim=2048,
        **model_args
    ):
        super(TimmBackbone, self).__init__()
        self.net = create_model(model_type, **model_args)

        layer = input_layer_getter(self.net); device = layer.weight.device
        if modify_input_channels and modify_input_channels!=layer.in_channels:
            layer.in_channels = modify_input_channels
            shape = layer.weight.shape; shape = torch.Size([shape[0],modify_input_channels,shape[2],shape[3]])
            layer.weight = nn.Parameter(torch.zeros(shape)).to(device)
            nninit.xavier_uniform_(layer.weight)
        layer = input_layer_getter(self.net)
        self.in_channels = layer.in_channels
        
        self.net = TV_U.IntermediateLayerGetter(self.net, return_layers={output_layer_name:"out"})
        self.net._modules[output_layer_name] = nn.Linear(self.net._modules[output_layer_name].in_features,embedding_dim,bias=True)
        self.net.to(device)

    def forward(self, data):
        return self.net(data)['out']

class FUNC(nn.Module):
    def __init__(
        self,
        func_layer=None,
        norm_layer=None,
        activation=None,
        dropout:Optional[int]=None,
        dropout_inplace:bool=True,
        reserve_identity:bool=True,
    ):
        super(FUNC, self).__init__()
        self.layers = nn.Sequential(*(
            ([nn.Dropout(p=dropout,inplace=dropout_inplace)] if dropout is not None else ([nn.Identity()] if reserve_identity else []))
        +   ([func_layer] if func_layer is not None else ([nn.Identity()] if reserve_identity else []))
        +   ([norm_layer] if norm_layer is not None else ([nn.Identity()] if reserve_identity else []))
        +   ([activation] if activation is not None else ([nn.Identity()] if reserve_identity else []))
        ))
    
    def forward(self, data):
        return self.layers(data)

class FC(nn.Module):
    def __init__(
        self,
        input_dim:int,
        output_dim:int,
        bias=True,
        norm_layer=None,
        activation=None,
        dropout:Optional[int]=None,
        dropout_inplace:bool=True,
        flatten=True,
        reserve_identity:bool=True,
    ):
        super(FC, self).__init__()
        self.input_dim = input_dim; self.output_dim = output_dim
        self.layers = nn.Sequential(*(
            ([nn.Dropout(p=dropout,inplace=dropout_inplace)] if dropout is not None else ([nn.Identity()] if reserve_identity else []))
        +   ([nn.Linear(input_dim,output_dim,bias=bias)])
        +   ([norm_layer] if norm_layer is not None else ([nn.Identity()] if reserve_identity else []))
        +   ([activation] if activation is not None else ([nn.Identity()] if reserve_identity else []))
        ))
        self.flatten = flatten
    
    def forward(self, data):
        return self.layers(torch.flatten(data,start_dim=1)) if self.flatten else self.layers(data)

class CONV(nn.Module):
    def __init__(
        self,
        in_channels:int,
        out_channels:int,
        kernel_size:int,
        stride:int=1,
        padding:int=0,
        dilation:int=1,
        groups:int=1,
        padding_mode:str='zeros',
        bias:bool=True,
        norm_layer=None,
        activation=None,
        dropout:Optional[int]=None,
        dropout_inplace:bool=True,
        conv1d:bool=False,
        reserve_identity:bool=True,
    ):
        super(CONV, self).__init__()
        self.in_channels = in_channels; self.out_channels = out_channels
        self.layers = nn.Sequential(*(
            ([nn.Dropout(p=dropout,inplace=dropout_inplace)] if dropout is not None else ([nn.Identity()] if reserve_identity else []))
        +   ([(nn.Conv1d if conv1d else nn.Conv2d)(in_channels=in_channels,out_channels=out_channels,\
                                                   kernel_size=kernel_size,stride=stride,padding=padding,\
                                                   dilation=dilation,groups=groups,padding_mode=padding_mode,bias=bias)])
        +   ([norm_layer] if norm_layer is not None else ([nn.Identity()] if reserve_identity else []))
        +   ([activation] if activation is not None else ([nn.Identity()] if reserve_identity else []))
        ))
    
    def forward(self, data):
        return self.layers(data)


from .file_utils import *
from PIL import Image
import numpy as np

def GrayScale(r, g, b):
    """Fast computation of gray scale given red, green, and blue data, each could be a value or an array of arbitrary shape, as long as the shape is the same.

    The computation is not accurate, instead, it uses fixed finite 1e-3 precision. The reason is to speedup computation using only integer:

    GrayScale(r,g,b) = floor((r * 299 + g * 587 + b * 114 + 500) / 1000)

    Args:
        r: The red channel of the image.
        g: The green channel of the image.
        b: The blue channel of the image.
    Returns:
        Any: The gray scale computed from r, g, b data, with finite 1e-3 precision.
    """
    r = np.array(r, dtype=int); g = np.array(g, dtype=int); b = np.array(b, dtype=int)
    return np.array(np.floor((r * 299 + g * 587 + b * 114 + 500) / 1000), dtype=np.uint8)

def DataToImg(data):
    """Convert an array-like data (values in 0 ~ 255) to image.

    Args:
        data: The array-like data.
    Returns:
        Image: The converted image.
    """
    return data if isinstance(data, Image.Image) else Image.fromarray(np.uint8(data))

def d2i(data):
    """Convert an array-like data (values in 0 ~ 255) to image.

    This is a short alias for `DataToImg`.

    Args:
        data: The array-like data.
    Returns:
        Image: The converted image.
    """
    return data if isinstance(data, Image.Image) else Image.fromarray(np.uint8(data))

def ImgToData(data):
    """Convert a PIL Image to array-like data.

    Args:
        img: The PIL Image.
    Returns:
        np.ndarray: The converted data.
    """
    return data if isinstance(data, np.ndarray) else np.array(data)

def i2d(data):
    """Convert a PIL Image to array-like data.

    This is a short alias for `ImgToData`.

    Args:
        img: The PIL Image.
    Returns:
        np.ndarray: The converted data.
    """
    return data if isinstance(data, np.ndarray) else np.array(data)

def DebugImgData(data):
    """Show an array-like data (values in 0 ~ 255) as image.

    Args:
        data: The array-like data.
    Returns:
        None
    """
    img = d2i(data); img.show(); img.close()

def SaveImgData(data, path, compress_level=0):
    """Save an array-like data (values in 0 ~ 255) as image, with no image compression.

    Args:
        data: The array-like data.
        path: The target image path.
        compress_level (bool): The `compress_level` of `Image.save`.
    Returns:
        None
    """
    img = d2i(data); img.save(p2s(path), compress_level=compress_level); img.close()

def RGBImg(img, transparent=np.array([0,0,0],dtype=float)):
    """Convert an image to image of 3 channels. Assuming the input image are RGB if it is already a 3-channel image, and RGBA if it is a 4-channel image.

    Supported image shapes are: (W,H), (W,H,1), (W,H,3), (W,H,4).

    Args:
        img: The image to be converted.
        transparent: The transparent color in RGB form (used for converting RGBA), e.g.: `np.array([255,255,255],dtype=float)` stands for white.
    Returns:
        np.ndarray: The converted image data.
    """
    img = deepcopy(i2d(img))
    if len(img.shape) == 2:
        img = np.stack([img, img, img], axis=(-1))
    elif img.shape[(-1)] == 1:
        img = np.concatenate([img, img, img], axis=(-1))
    elif img.shape[(-1)] == 3:
        pass
    elif img.shape[(-1)] == 4:
        img = img.astype(np.float); bg = np.broadcast_to(transparent,img[:,:,:3].shape)
        alpha = np.stack([img[:, :, 3], img[:, :, 3], img[:, :, 3]], axis=(-1)) / 255
        img = np.array((np.uint8((img[:, :, :3]*alpha+bg*(1-alpha)))), dtype=(np.uint8))
    else:
        raise SystemError
    return img


from .file_utils import *
from .args_utils import MemberDict
from .misc_utils import Clipped
import PySide6

from PySide6 import QtWidgets
from PySide6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QDialog,
    QFileDialog,
    QFrame,
    QGraphicsDropShadowEffect,
    QGridLayout,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QLayout,
    QLineEdit,
    QListWidget,
    QMainWindow,
    QMessageBox,
    QProgressBar,
    QProgressDialog,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QSpinBox,
    QSpinBox,
    QStackedWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from PySide6 import QtGui
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGuiApplication,
    QIcon,
    QLinearGradient,
    QPalette,
    QPainter,
    QPixmap,
    QRadialGradient
)

from PySide6 import QtCore
from PySide6.QtCore import (
    QCoreApplication,
    QTimer,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QUrl,
    Qt,
)

from functools import partial

def QWrapped(widget):
    grid = QGridLayout()
    widget.setLayout(grid)
    widget.setMaximumSize(65536, 65536)
    widget.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
    grid.setContentsMargins(10, 10, 10, 10)
    grid.setSpacing(2)
    return widget

def QAdd(widget, module, pos, span, **args):
    instance = module(widget, **args)
    widget.layout().addWidget(instance, pos[0], pos[1], span[0], span[1])
    instance.setMaximumSize(65536, 65536)
    instance.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
    return instance

def QGetInput(base, title, question, default=None):
    answer, done = QInputDialog.getText(base, title, question)
    return answer if done else default

def QGetFolder(base, title, default=""):
    folder = QFileDialog.getExistingDirectory(base, title, default)
    return folder if folder!="" and ExistFolder(folder) else None

def QGetFile(base, title, default="C:/", filter="All Files (*)"):
    file, file_type = QFileDialog.getOpenFileName(base, title, default, filter=filter)
    return file if file!="" and file_type!="" and ExistFile(file) else None

def QGetNewFile(base, title, default="C:/", filter="All Files (*)"):
    file, file_type = QFileDialog.getSaveFileName(base, title, default, filter=filter)
    return file if file!="" and file_type!="" else None

def QHint(base, title, message):
    hint = QMessageBox(base)
    hint.setWindowTitle(title)
    hint.setWindowModality(Qt.WindowModal)
    hint.setText(message)
    hint.exec()

def QConfirm(base, title, message):
    return (QMessageBox.question(base, title, message, QMessageBox.Yes | QMessageBox.No, QMessageBox.No) == QMessageBox.Yes)

# Modified from https://github.com/Wanderson-Magalhaes/Python_PySide2_Circular_ProgressBar_Modern_GUI
class Ui_SplashScreen(object):
    def setupUi(self, SplashScreen, **args):
        if SplashScreen.objectName():
            SplashScreen.setObjectName(u"SplashScreen")
        SplashScreen.resize(360, 360)
        self.centralwidget = QWidget(SplashScreen)
        self.centralwidget.setObjectName(u"centralwidget")
        self.circularProgressBarBase = QFrame(self.centralwidget)
        self.circularProgressBarBase.setObjectName(u"circularProgressBarBase")
        self.circularProgressBarBase.setGeometry(QRect(10, 10, 320, 320))
        self.circularProgressBarBase.setFrameShape(QFrame.NoFrame)
        self.circularProgressBarBase.setFrameShadow(QFrame.Raised)
        self.circularProgress = QFrame(self.circularProgressBarBase)
        self.circularProgress.setObjectName(u"circularProgress")
        self.circularProgress.setGeometry(QRect(10, 10, 300, 300))
        self.circularProgress.setStyleSheet(u"QFrame{\n"
"   border-radius: 150px;\n"
"   background-color: qconicalgradient(cx:0.5, cy:0.5, angle:90, stop:0.749 rgba(255, 0, 127, 0), stop:0.750 rgba(85, 170, 255, 255));\n"
"}")
        self.circularProgress.setFrameShape(QFrame.NoFrame)
        self.circularProgress.setFrameShadow(QFrame.Raised)
        self.circularBg = QFrame(self.circularProgressBarBase)
        self.circularBg.setObjectName(u"circularBg")
        self.circularBg.setGeometry(QRect(10, 10, 300, 300))
        self.circularBg.setStyleSheet(u"QFrame{\n"
"   border-radius: 150px;\n"
"   background-color: "+args['bg_color']+";\n"
"}")
        self.circularBg.setFrameShape(QFrame.NoFrame)
        self.circularBg.setFrameShadow(QFrame.Raised)
        self.container = QFrame(self.circularProgressBarBase)
        self.container.setObjectName(u"container")
        self.container.setGeometry(QRect(20, 20, 280, 280))
        self.container.setStyleSheet(u"QFrame{\n"
"   border-radius: 135px;\n"
"   background-color: "+args['container_color']+";\n"
"}")
        self.container.setFrameShape(QFrame.NoFrame)
        self.container.setFrameShadow(QFrame.Raised)
        self.widget = QWidget(self.container)
        self.widget.setObjectName(u"widget")
        self.widget.setGeometry(QRect(30, 40, 220, 210))
        self.gridLayout = QGridLayout(self.widget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.labelTitle = QLabel(self.widget)
        self.labelTitle.setObjectName(u"labelTitle")
        font = QFont()
        font.setFamily(args['title_font_family'])
        font.setPointSize(args['title_font_size'])
        self.labelTitle.setFont(font)
        self.labelTitle.setStyleSheet(u"background-color: none;\n"
"color: #FFFFFF")
        self.labelTitle.setAlignment(Qt.AlignCenter)

        self.gridLayout.addWidget(self.labelTitle, 0, 0, 1, 1)

        self.labelPercentage = QLabel(self.widget)
        self.labelPercentage.setObjectName(u"labelPercentage")
        font1 = QFont()
        font1.setFamily(args['percentage_font_family'])
        font1.setPointSize(args['percentage_font_size'])
        self.labelPercentage.setFont(font1)
        self.labelPercentage.setStyleSheet(u"background-color: none;\n"
"color: #FFFFFF")
        self.labelPercentage.setAlignment(Qt.AlignCenter)

        self.gridLayout.addWidget(self.labelPercentage, 1, 0, 1, 1)

        self.labelLoadingInfo = QLabel(self.widget)
        self.labelLoadingInfo.setObjectName(u"labelLoadingInfo")
        self.labelLoadingInfo.setMinimumSize(QSize(0, 24))
        self.labelLoadingInfo.setMaximumSize(QSize(16777215, 24))
        font2 = QFont()
        font2.setFamily(args['info_font_family'])
        font2.setPointSize(args['info_font_size'])
        self.labelLoadingInfo.setFont(font2)
        self.labelLoadingInfo.setStyleSheet(u"QLabel{\n"
"   border-radius: 10px;    \n"
"   background-color: "+args['label_color']+";\n"
"   color: #FFFFFF;\n"
"   margin-left: 0px;\n"
"   margin-right: 0px;\n"
"}")
        self.labelLoadingInfo.setFrameShape(QFrame.NoFrame)
        self.labelLoadingInfo.setAlignment(Qt.AlignCenter)

        self.gridLayout.addWidget(self.labelLoadingInfo, 2, 0, 1, 1)

        self.labelCredits = QLabel(self.widget)
        self.labelCredits.setObjectName(u"labelCredits")
        self.labelCredits.setFont(font2)
        self.labelCredits.setStyleSheet(u"background-color: none;\n"
"color: rgb(155, 155, 255);")
        self.labelCredits.setAlignment(Qt.AlignCenter)

        self.gridLayout.addWidget(self.labelCredits, 3, 0, 1, 1)

        self.circularBg.raise_()
        self.circularProgress.raise_()
        self.container.raise_()
        layout = QVBoxLayout()
        layout.addWidget(self.centralwidget)
        SplashScreen.setLayout(layout)

        self.retranslateUi(SplashScreen, **args)

        QMetaObject.connectSlotsByName(SplashScreen)

    def retranslateUi(self, SplashScreen, **args):
        SplashScreen.setWindowTitle(QCoreApplication.translate("SplashScreen", args['window_name'], None))
        self.labelTitle.setText(QCoreApplication.translate("SplashScreen", u"<html><head/><body><p><span style=\" font-weight:bold; color:#9b9bff;\">"+args['title']+"</span> "+args['subtitle']+"</p></body></html>", None))
        self.labelPercentage.setText(QCoreApplication.translate("SplashScreen", u"<p><span style=\" font-size:"+str(args["percentage_font_size"])+"pt;\">"+str(args['__expression__'])+"</span></p>", None))
        self.labelLoadingInfo.setText(QCoreApplication.translate("SplashScreen", args['__loading_info__'], None))
        self.labelCredits.setText(QCoreApplication.translate("SplashScreen", args['credits'], None))

class QCircularProgressDialog(QDialog):
    def __init__(self, refresh=100, cut=32, optimize=False, window_name="", title="", subtitle="", loading_info="loading...", credits="by: Wanderson M. Pimenta"):
        super().__init__()
        self.cut = max(cut,3)
        self.refresh = refresh
        self.optimize = optimize
        self.ui = Ui_SplashScreen()
        self.ui_info = MemberDict({
            'value': 0,

            'window_name': "",
            'title': title,
            'subtitle': subtitle,
            'loading_info': loading_info,
            '__loading_info__': None,
            'credits': credits,

            'title_font_family': "YaHei Consolas Hybrid",
            'title_font_size': 10,
            'info_font_size': 8,
            'info_font_family': "YaHei Consolas Hybrid",
            'percentage_font_size': 32,
            'percentage_font_family': "Consolas",

            'progress_color': "rgba(85, 170, 255, 255)",
            'container_color': "rgb(77, 77, 127)",
            'bg_color': "rgba(77, 77, 127, 120)",
            'label_color': "rgb(93, 93, 154)",

            'expression': None,
            '__expression__': None
        })
        self.ui.setupUi(self, **self.ui_info)

        self.setWindowFlags(QtCore.Qt.FramelessWindowHint)
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)

        self.shadow = QGraphicsDropShadowEffect(self)
        self.shadow.setBlurRadius(20)
        self.shadow.setXOffset(0)
        self.shadow.setYOffset(0)
        self.shadow.setColor(QColor(0, 0, 0, 120))
        self.ui.circularBg.setGraphicsEffect(self.shadow)

        self.finished = False
        self.timer = QTimer()
        self.timer.timeout.connect(self._refresh)
        self.timer.start(self.refresh)
        self.setValue(0)

        self.center()
        self.show()
        self.activateWindow()

    def center(self):
        qr = self.frameGeometry()
        cp = QGuiApplication.primaryScreen().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def _refresh(self):
        QApplication.processEvents()

    def setUiInfo(self, infos, refresh=False):
        self.ui_info.update(infos)
        if self.ui_info.expression is None:
            self.ui_info.__expression__ = f"{self.ui_info.value}%"
        else:
            self.ui_info.__expression__ = self.ui_info.expression
        if self.ui_info.loading_info is None:
            self.ui_info.__loading_info__ = None
        else:
            self.ui_info.__loading_info__ = self.ui_info.loading_info if len(self.ui_info.loading_info)<self.cut else "..."+self.ui_info.loading_info[-self.cut+3:]
        if self.finished:
            return
        if self.ui_info.value >= 100:
            self.finished = True
            self.timer.stop()
            self.close()
        else:
            progress = (100 - self.ui_info.value) / 100.0
            stop_1 = str(progress - 0.001); stop_2 = str(progress)
            styleSheet = """
            QFrame{
                border-radius: 150px;
                background-color: qconicalgradient(cx:0.5, cy:0.5, angle:90, stop:{STOP_1} rgba(255, 0, 127, 0), stop:{STOP_2} {COLOR});
            }
            """
            newStylesheet = styleSheet.replace("{STOP_1}", stop_1).replace("{STOP_2}", stop_2).replace('{COLOR}', self.ui_info.progress_color)
            self.ui.circularProgress.setStyleSheet(newStylesheet); self.ui.retranslateUi(self, **self.ui_info); QApplication.processEvents()

    def setValue(self, value):
        self.setUiInfo({'value':Clipped(int(value), 0, 100)})

    def terminate(self):
        self.finished = True
        self.timer.stop()
        self.close()




from .file_utils import *
from .serialize_utils import DumpsJson, LoadJson
import argparse

class MemberDict(dict):
    """A dict object whose items can be accessed using `.`.
    """
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def view(self, backend='json', indent=4):
        return DumpsJson(dict(self),backend=backend,indent=indent)
    
    def __add__(self, b):
        d = MemberDict(self)
        for key in b:
            if key in d:
                d[key] += b[key]
            else:
                d[key] = b[key]
        return d
    
    def __or__(self, b):
        d = MemberDict(self)
        for key in b:
            d[key] = b[key]
        return d
    
    def __mul__(self, b):
        d = MemberDict(self)
        for key in d:
            d[key] *= b
        return d

    def __div__(self, b):
        d = MemberDict(self)
        for key in d:
            d[key] /= b
        return d

    def __getstate__(self):
        return dict(self)

    def __setstate__(self, state):
        self.__dict__ = state

class ArgumentDescriptor(object):
    """The base descriptor that wraps arguments of `add_argument` for `argparse.ArgumentParser`.
    """
    def __init__(self, dest:str, full:str=None, short:str=None, required:bool=False, **kwargs):
        """The following are different from `add_argument`:
        1. '-' in `dest` will be replaced to '_'.
        2. All unnecessary arguments are represented as both short name (i.e.: `-`) and full name (i.e.: `--`), the full name is  `dest` ('_' will be replaced to '-') by default, and the short name is the first character of full name by default. Please specify `short` manually when conflict occurs. 
        3. All necessary arguments are represented with `dest` ('-' will be replaced to '_'). Arguments are determined necessary if and only if `required` is set to True.
        
        Args:
            dest (str): The `dest` arg of `add_argument`.
            full (str): The full name (i.e.: `--`) of the argument.
            short (str): The short name (i.e.: `-`) of the argugment.
            required (bool): If True, the argument is necessary, otherwise unnecessary.
        """
        dest = dest.replace('-','_')
        if full is None:
            full = dest.replace('_','-')
        if short is None:
            short = full[0]
        short = "-"+short; full = "--"+full
        self.name = (dest, ) if required else (short, full)
        self.args = {} if required else {'dest':dest}
        self.args.update(kwargs)

    def register(self, parser):
        """Register the argument described by the descriptor to `parset`.

        Args:
            parser: The parser object that this argument will be registered to.
        """
        parser.add_argument(
            *self.name,
            **self.args
        )

class IntArgumentDescriptor(ArgumentDescriptor):
    """The descriptor for arguments of `type` equal to `int`.
    """
    def __init__(self, dest:str, full:str=None, short:str=None, required:bool=False, **kwargs):
        """The following are different from `add_argument`:
        1. '-' in `dest` will be replaced to '_'.
        2. All unnecessary arguments are represented as both short name (i.e.: `-`) and full name (i.e.: `--`), the full name is  `dest` ('_' will be replaced to '-') by default, and the short name is the first character of full name by default. Please specify `short` manually when conflict occurs. 
        3. All necessary arguments are represented with `dest` ('-' will be replaced to '_'). Arguments are determined necessary if and only if `required` is set to True.
        4. The `type` argument will be set to `int` by default (and `type` specifier in `kwargs` will be override).

        Args:
            dest (str): The `dest` arg of `add_argument`.
            full (str): The full name (i.e.: `--`) of the argugment.
            short (str): The short name (i.e.: `-`) of the argugment.
            required (bool): If True, the argument is necessary, otherwise unnecessary.
        """
        kwargs['type'] = int; super(IntArgumentDescriptor, self).__init__(dest=dest, full=full, short=short, required=required, **kwargs)

class StrArgumentDescriptor(ArgumentDescriptor):
    """The descriptor for arguments of `type` equal to `str`.
    """
    def __init__(self, dest:str, full:str=None, short:str=None, required:bool=False, **kwargs):
        """The following are different from `add_argument`:
        1. '-' in `dest` will be replaced to '_'.
        2. All unnecessary arguments are represented as both short name (i.e.: `-`) and full name (i.e.: `--`), the full name is  `dest` ('_' will be replaced to '-') by default, and the short name is the first character of full name by default. Please specify `short` manually when conflict occurs. 
        3. All necessary arguments are represented with `dest` ('-' will be replaced to '_'). Arguments are determined necessary if and only if `required` is set to True.
        4. The `type` argument will be set to `str` by default (and `type` specifier in `kwargs` will be override).

        Args:
            dest (str): The `dest` arg of `add_argument`.
            full (str): The full name (i.e.: `--`) of the argugment.
            short (str): The short name (i.e.: `-`) of the argugment.
            required (bool): If True, the argument is necessary, otherwise unnecessary.
        """
        kwargs['type'] = str; super(StrArgumentDescriptor, self).__init__(dest=dest, full=full, short=short, required=required, **kwargs)

class BoolArgumentDescriptor(ArgumentDescriptor):
    """The descriptor for arguments of `type` equal to `bool`.
    """
    def __init__(self, dest:str, full:str=None, short:str=None, required:bool=False, **kwargs):
        """The following are different from `add_argument`:
        1. '-' in `dest` will be replaced to '_'.
        2. All unnecessary arguments are represented as both short name (i.e.: `-`) and full name (i.e.: `--`), the full name is  `dest` ('_' will be replaced to '-') by default, and the short name is the first character of full name by default. Please specify `short` manually when conflict occurs. 
        3. All necessary arguments are represented with `dest` ('-' will be replaced to '_'). Arguments are determined necessary if and only if `required` is set to True.
        4. The `type` argument will be set to `bool` by default (and `type` specifier in `kwargs` will be override).

        Args:
            dest (str): The `dest` arg of `add_argument`.
            full (str): The full name (i.e.: `--`) of the argugment.
            short (str): The short name (i.e.: `-`) of the argugment.
            required (bool): If True, the argument is necessary, otherwise unnecessary.
        """
        kwargs['type'] = bool; super(BoolArgumentDescriptor, self).__init__(dest=dest, full=full, short=short, required=required, **kwargs)

class FloatArgumentDescriptor(ArgumentDescriptor):
    """The descriptor for arguments of `type` equal to `float`.
    """
    def __init__(self, dest:str, full:str=None, short:str=None, required:bool=False, **kwargs):
        """The following are different from `add_argument`:
        1. '-' in `dest` will be replaced to '_'.
        2. All unnecessary arguments are represented as both short name (i.e.: `-`) and full name (i.e.: `--`), the full name is  `dest` ('_' will be replaced to '-') by default, and the short name is the first character of full name by default. Please specify `short` manually when conflict occurs. 
        3. All necessary arguments are represented with `dest` ('-' will be replaced to '_'). Arguments are determined necessary if and only if `required` is set to True.
        4. The `type` argument will be set to `float` by default (and `type` specifier in `kwargs` will be override).

        Args:
            dest (str): The `dest` arg of `add_argument`.
            full (str): The full name (i.e.: `--`) of the argugment.
            short (str): The short name (i.e.: `-`) of the argugment.
            required (bool): If True, the argument is necessary, otherwise unnecessary.
        """
        kwargs['type'] = float; super(FloatArgumentDescriptor, self).__init__(dest=dest, full=full, short=short, required=required, **kwargs)

class SwitchArgumentDescriptor(ArgumentDescriptor):
    """The descriptor for arguments of `action` equal to `store_true` or `store_false`.
    """
    def __init__(self, dest:str, to:bool=True, full:str=None, short:str=None, required:bool=False, **kwargs):
        """The following are different from `add_argument`:
        1. '-' in `dest` will be replaced to '_'.
        2. All unnecessary arguments are represented as both short name (i.e.: `-`) and full name (i.e.: `--`), the full name is  `dest` ('_' will be replaced to '-') by default, and the short name is the first character of full name by default. Please specify `short` manually when conflict occurs. 
        3. All necessary arguments are represented with `dest` ('-' will be replaced to '_'). Arguments are determined necessary if and only if `required` is set to True.
        4. The `action` argument will be set to 'store_true'/'store_false' by default (and `action` specifier in `kwargs` will be override).

        Args:
            dest (str): The `dest` arg of `add_argument`.
            to (bool): If True, the `action` is set to 'store_true', otherwise 'store_false'.  
            full (str): The full name (i.e.: `--`) of the argugment.
            short (str): The short name (i.e.: `-`) of the argugment.
            required (bool): If True, the argument is necessary, otherwise unnecessary.
        """
        kwargs['action'] = 'store_'+['false','true'][to]; super(SwitchArgumentDescriptor, self).__init__(dest=dest, full=full, short=short, required=required, **kwargs)

class ConstArgumentDescriptor(ArgumentDescriptor):
    """The descriptor for arguments of `action` equal to `store_const`.
    """
    def __init__(self, dest:str, const, full:str=None, short:str=None, required:bool=False, **kwargs):
        """The following are different from `add_argument`:
        1. '-' in `dest` will be replaced to '_'.
        2. All unnecessary arguments are represented as both short name (i.e.: `-`) and full name (i.e.: `--`), the full name is  `dest` ('_' will be replaced to '-') by default, and the short name is the first character of full name by default. Please specify `short` manually when conflict occurs. 
        3. All necessary arguments are represented with `dest` ('-' will be replaced to '_'). Arguments are determined necessary if and only if `required` is set to True.
        4. The `action` argument will be set to 'store_const' and the `const` argument will be set to `const` by default (and `action` and `const` specifier in `kwargs` will be override).

        Args:
            dest (str): The `dest` arg of `add_argument`.
            const: The const value to be stored.  
            full (str): The full name (i.e.: `--`) of the argugment.
            short (str): The short name (i.e.: `-`) of the argugment.
            required (bool): If True, the argument is necessary, otherwise unnecessary.
        """
        kwargs['action'] = 'store_const'; kwargs['const'] = const; super(ConstArgumentDescriptor, self).__init__(dest=dest, full=full, short=short, required=required, **kwargs)

class LiteralArgumentDescriptor(ArgumentDescriptor):
    """The descriptor for arguments with finite `choices`.
    """
    def __init__(self, dest:str, choices:List, full:str=None, short:str=None, required:bool=False, **kwargs):
        """The following are different from `add_argument`:
        1. '-' in `dest` will be replaced to '_'.
        2. All unnecessary arguments are represented as both short name (i.e.: `-`) and full name (i.e.: `--`), the full name is  `dest` ('_' will be replaced to '-') by default, and the short name is the first character of full name by default. Please specify `short` manually when conflict occurs. 
        3. All necessary arguments are represented with `dest` ('-' will be replaced to '_'). Arguments are determined necessary if and only if `required` is set to True.
        4. The `choices` argument will be set to `choices` by default (and `choices` specifier in `kwargs` will be override).

        Args:
            dest (str): The `dest` arg of `add_argument`.
            choices (List): The options specified for the argument.  
            full (str): The full name (i.e.: `--`) of the argugment.
            short (str): The short name (i.e.: `-`) of the argugment.
            required (bool): If True, the argument is necessary, otherwise unnecessary.
        """
        kwargs['choices'] = choices; super(LiteralArgumentDescriptor, self).__init__(dest=dest, full=full, short=short, required=required, **kwargs)

class ListArgumentDescriptor(ArgumentDescriptor):
    """The descriptor for arguments of `action` equal to `extend`.
    """
    def __init__(self, dest:str, full:str=None, short:str=None, required:bool=False, **kwargs):
        """The following are different from `add_argument`:
        1. '-' in `dest` will be replaced to '_'.
        2. All unnecessary arguments are represented as both short name (i.e.: `-`) and full name (i.e.: `--`), the full name is  `dest` ('_' will be replaced to '-') by default, and the short name is the first character of full name by default. Please specify `short` manually when conflict occurs. 
        3. All necessary arguments are represented with `dest` ('-' will be replaced to '_'). Arguments are determined necessary if and only if `required` is set to True.
        4. The `type` argument will be set to `int` by default (and `type` specifier in `kwargs` will be override).

        Args:
            dest (str): The `dest` arg of `add_argument`.
            full (str): The full name (i.e.: `--`) of the argugment.
            short (str): The short name (i.e.: `-`) of the argugment.
            required (bool): If True, the argument is necessary, otherwise unnecessary.
        """
        kwargs['action'] = 'extend'; kwargs['nargs'] = kwargs['nargs'] if 'nargs' in kwargs else '*'; kwargs['default'] = kwargs['default'] if 'default' in kwargs else list()
        super(ListArgumentDescriptor, self).__init__(dest=dest, full=full, short=short, required=required, **kwargs)

class HeavenArguments(MemberDict):
    """An argument dict extended from `MemberDict`.
    """
    def __iter__(self):
        for key in sorted(self.keys()):
            yield key, self[key]

    def __str__(self):
        return DumpsJson(dict(self),indent=4)

    @classmethod
    def from_parser(cls, descriptors:List, **parser_args):
        """Construct arguments from a list of argument descriptors and `argparse`.

        Notice that this function will construct argparser and parse all input args.

        Args:
            descriptors (List): The list of argument descriptors.
            parser_args: Arguments for constructing `argparse.ArgumentParser`.
        Returns:
            Arguments: The parsed arguments.
        """
        parser = argparse.ArgumentParser(**parser_args)
        for descriptor in descriptors:
            descriptor.register(parser)
        return cls(vars(parser.parse_args()))
    
    @classmethod
    def from_json(cls, path, backend='json'):
        """Construct arguments from a json file.

        Args:
            path: The json file path.
            backend (str): Specify backend for saving an object in str format. Please refer to function `BUILTIN_JSON_BACKENDS()` for built-in backends.
        Returns:
            Arguments: The parsed arguments.
        """
        return cls(LoadJson(path,backend=backend))
    
    @classmethod
    def from_object(cls, object):
        """Construct arguments from an object.

        Args:
            obj: The object.
        Returns:
            Arguments: The parsed arguments.
        """
        return cls(vars(object))
    
    @classmethod
    def from_properties(cls, path, errors:Literal['interupt','strict','ignore']='interupt'):
        """Construct arguments from a properties file.

        Args:
            path: The properties file path.
            errors (str): Specify behavior when encountering errors.
                          If 'interupt', an Exception will be raised immediately at the error line.
                          If 'strict', a ValueError will be raised to report all error lines.
                          If 'ignore', all error lines will be skipped.
        Returns:
            Arguments: The parsed arguments.
        """
        args = {}; failures = []
        with open(path,"r") as f:
            i = 0
            for line in f:
                try:
                    if line.strip()!='':
                        property = line.strip().split('#')[0]
                        key, value = property.strip().split(' ')
                        args.update(key.strip(),value.strip())
                except Exception as e:
                    if errors=="interupt":
                        raise e
                    failures.append((i,line,e))
                i += 1
        if errors=="strict":
            raise ValueError("\n".join(failures))
        return cls(args)